from nzshm_common.grids.io import latlon_from_base64_zip


class NZ01nb1v0:
    nz = None

    def load(self):
        if self.nz is None:
            self.nz = latlon_from_base64_zip(data)
        return self.nz


data = (
    "UEsDBD8AAgAOAHJZ81Q6lt94aQYAAJipAAAEAAAAZGF0YQkEBQBdAACAAAAY"
    "jaxcNpfygoevllxFWkF0RZzJXrt41v/3qEQcJILW2eVT9YyfGk8NRy8Z/+c6"
    "yMlji9f6EFImZ94p8AMP4CQk+TdFX+Jy+ruB6qW7Mng+qXtIGBBV+eznCl7h"
    "SLc2RaDifGmN+MrsGrRVK04cKItx6R7B4MiwHJUEU/jTCVrKFPFhzL+Gh4aN"
    "b+y6YtnIpO3j7eCDnFckPgxDblz10tJQVA9kJwCemvFLdeQAHou5BKe0Ohs1"
    "gdGrurND82Hz2v9COaxV1CibI0EIBt44DizqMlCdEo+oaCmL34dha8ePedJv"
    "N6cUCtAYM344A8QbG3D7k7DqeSFaBD5u2dY+Qcj7tNCROjdXr1W6fr6jD5ty"
    "udnR12eoOe+WYgh7Qsm/bjTFIPEBvhyUSfVzp8MYqViSRIHg6DnGgQeQDA9f"
    "/hxNPprmWRY8YQXtU5+54QTDsBfnMVzPlIiF73T3Ap3PV6ZvRSXXyJh6SYjE"
    "6pMHpvYcbWjHrNSptF76HH4XHQOm+PFmZz8AOBVcWdfRb7dpCCZ8LJS8anfK"
    "zmVvoq+Sha+GNZlcQAUdJSe6++7dTvcIrhFIe6bpnMiq1qQ7CJHB74l6DTTE"
    "ZAez0+uuGbfxQGi5QqeEwMVxW0bWQ52BQ1yjLW3mFmJJl99Y7FmyrL28OLEx"
    "Uhn01iquVwMEXajYUBcPic0euzezxP3iFqs9Y46H4BtHdMj3ItJ2j2GeqEx9"
    "JSVl/MUaL7GijhKkmky2DWMk3b5z3bKVQHefSHrC5ywPE1j0n2jVOzI4hy+s"
    "LajXv192DtOnZ4t6+L3xMSY9lgk4rqWCdrV+iD8bBeMM0v8IppOupdUTYlq2"
    "18vNZuiG1SuCz6ShQ8rYQw1/5NPwSrGewxkY3VBUWLh8m1RKIpauN+5eAfQH"
    "XEyIPWs4EUZDZmOsYijwtYcgPXfzuonwOihddtKV6+Zyu5JnoPuQZlKLT0Zr"
    "OSk4KX9phi4hBYRPBzEKu/XoqeTUbuc6i3B20h6lAza7EiNfO6Xux51jPcNc"
    "Y/48WujCncBa9gdLRdKcpsVyKqU9P1hLF2vc4mJqP+sqPQ+Wru9/T407URso"
    "6iBU+br604K4/v+B7kTEgPog9wdPhz8eKY7huKrcgK5OLdExxU2o/D+W1xrZ"
    "zAUNVWrJCo5mGoVrhQXUUDmFu856+iaxGNkGRCfBrs2rtuekN7Z5VxYBocQk"
    "s+ixtHkGBKsM6lp6X9s1s7BOH2Pce0kOqsTD8oUhH3yI7tkpbh5N50jHP6n8"
    "ZK3eMyLdxUtvuz+pvHSW1dPfZJY40jYD+Qbbf/Jb59U/0HPKRwa6ir6V4fSZ"
    "c3MBYqibl1lZOVX2mjzW8BzH6oMJ6LsBS79xZN7cYmpxLvPLZPt8rhCTK5BF"
    "6S5yMumhYbz1v+rdlxT2ZiD3kuMsIWsYNwBHR/L6DmqUMmTpqm8jT3o0VOST"
    "hmK/D67P0ks8t8m/bC+xa0Ix0wmteBOOV1MZfpEWSH4s280G3LGwnSW0AUgS"
    "ZUo1SEqflOw4pEbP6mKA5HaaFzAK36Jh0TeMl71O/2mvGoMhEi2DVeYk0aqo"
    "4RSWNNAD4ws5FglwOJm+nbzPQ3/qAQu6LJY8ZIS9ySm6kY8FlXQqo13TE2gY"
    "fP0wHne5QunpQXC8pV1LsoxMF/rIRbVexHp62cUbOt4pLGwbiGTWVteoSTop"
    "MKE8UxuzrHo8Wc1hfNGuLl7eL8BL1kyels3goeAK5nNCKofMNsZ88UecM8Wv"
    "G9ZjPguK5RzZPx0PopnX38aKaQk86/tjBnCKTpoE24B2sJZO6Vd8xagyKdfP"
    "gPubkwIT5XSlC3AqvO/uHGigW72yt95xk5QRNyeCfrybwes7rBGWTUMox3WF"
    "Mbpijfc5hyvGGkDiZRy1G8IJQqnBqZEuze+bmXAwbq8HAEAYkgyXw9+jW4IB"
    "Zqj7dNynynmrRTs49evKR59eyqx/eRODpAMr2YNRDckY3tBk/57IzFuxLxwM"
    "SxoPNWfNbdtO/Ugv0jfOu9OHwKyyrYaiKC1eae6WyKzFNz3CUrE/kmOFOlKy"
    "mtdkggUPVTbMe+fX/TFURDZCk9SzHQMNzAbPl2Ay4PVzDDPp5jGn66zVkjOR"
    "fROw8qj9Wb88r29/sdhJibzZ0SYxr4W/lMd9Tg/tW0e11m5kku24OUENgKcA"
    "BOFu5sn/10lTKVBLAQI/Az8AAgAOAHJZ81Q6lt94aQYAAJipAAAEAAAAAAAA"
    "AAAAAACAAQAAAABkYXRhUEsFBgAAAAABAAEAMgAAAIsGAAAAAA=="
)



"""
Standard values that are useful in multiple places.
"""

DEFAULT_RESOLUTION = 0.001
"""
Default coordinate resolution in degrees.
This is typically used for
[`CodedLocation`](location/coded_location.md#nzshm_common.location.coded_location.CodedLocation) values.
"""



"""
This module contains constants and functions for referring to location or list of locations by an identifier.
"""

import csv
import importlib.resources as resources
import json
from collections import namedtuple
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

from nzshm_common.constants import DEFAULT_RESOLUTION
from nzshm_common.grids.region_grid import load_grid
from nzshm_common.location.coded_location import CodedLocation
from nzshm_common.location.types import LatLon

# Omitting country for now, focus on NZ
# https://service.unece.org/trade/locode/nz.htm

resource_dir = resources.files('nzshm_common.location.resources')

with resources.as_file(resource_dir / 'locations.json') as path:
    with path.open() as file:
        LOCATIONS = json.load(file)

with resources.as_file(resource_dir / 'nz_ids.json') as path:
    with path.open() as file:
        NZ_IDS = json.load(file)

LOCATIONS_BY_ID: Dict[str, Any] = {location["id"]: location for location in LOCATIONS}


LOCATION_LISTS = {
    "HB": {
        "id": "HB",
        "name": "Hawk's Bay high res grid with vs30",
        "locations": [loc["id"] for loc in LOCATIONS if "hb_" in loc["id"]],
    },
    "NZ": {"id": "NZ", "name": "Default NZ locations", "locations": NZ_IDS},
    "NZ2": {
        "id": "NZ2",
        "name": "Main Cities NZ",
        "locations": ["WLG", "CHC", "DUD", "NPL", "AKL", "ROT", "HLZ"],
    },
    "SRWG214": {
        "id": "SRWG214",
        "name": "Seismic Risk Working Group NZ code locations",
        # "locations": list(map(lambda idn: f"srg_{idn}", range(214))),
        "locations": [loc["id"] for loc in LOCATIONS if "srg_" in loc["id"]],
    },
    "ALL": {
        "id": "ALL",
        "name": "All locations",
        "locations": list(map(lambda loc: loc["id"], LOCATIONS)),
    },
}


def _get_macron_word_mapping() -> Dict[str, str]:
    """using the maori_names.csv file as received from LINZ rather than storing the mapping allows
    us to update without rebuilding the resource"""

    char_map_lower = {
        'ā': 'a',
        'ē': 'e',
        'ī': 'i',
        'ō': 'o',
        'ū': 'u',
    }
    char_map = {}
    for k, v in char_map_lower.items():
        char_map[k] = v
        char_map[k.upper()] = v.upper()
    translation_table = str.maketrans(char_map)

    word_mapping = dict()
    with resources.as_file(resource_dir / 'maori_names.csv') as path:
        with path.open(encoding='utf-8') as file:
            reader = csv.reader(file)
            _ = next(reader)
            for row in reader:
                name = row[1]  # second column of LINZ file contains names
                for word in name.split():  # treat each whole word seperatly
                    if any([char in char_map for char in word]):  # add to mapping if any characters have macron
                        word_nomacron = word.translate(translation_table)
                        word_mapping[word_nomacron] = word
    return word_mapping


WORD_MAPPING = _get_macron_word_mapping()


def get_name_with_macrons(name_input: str) -> str:
    """
    Corrects the spelling of Māori palce names by adding macrons. Place name spellings from
    LINZ "Place names of New Zealand".
    See https://www.linz.govt.nz/products-services/place-names/place-names-new-zealand
    and https://gazetteer.linz.govt.nz/maori_names.csv

    If the input name is not on the LINZ list, the function will return the input.

    Args:
        input_name: the name to correct with macrons

    Returns:
        the place name with the correct Māori spelling
    """

    words_in = name_input.split()
    return " ".join([_map_word(word) for word in words_in])


def _map_word(word_input):
    if word_output := WORD_MAPPING.get(word_input):
        return word_output
    return word_input


def _lat_lon(_id) -> Optional[LatLon]:
    loc = location_by_id(_id)
    if loc:
        return LatLon(loc['latitude'], loc['longitude'])
    return None


def _load_csv(locations_filepath, resolution):
    locs = []
    with locations_filepath.open('r') as locations_file:
        reader = csv.reader(locations_file)
        Location = namedtuple("Location", next(reader), rename=True)
        for row in reader:
            location = Location(*row)
            locs.append(CodedLocation(lat=float(location.lat), lon=float(location.lon), resolution=resolution))
    return locs


def location_by_id(location_code: str) -> Optional[Dict[str, Any]]:
    """
    Get the information for a location identified by an id.

    Parameters:
        location_code: the code (e.g. "WLG") for the location

    Returns:
        coded location of location_code

    Examples:
        >>> location_by_id("WLG")
        {'id': 'WLG', 'name': 'Wellington', 'latitude': -41.3, 'longitude': 174.78}
    """
    return LOCATIONS_BY_ID.get(location_code)


def get_locations(locations: Iterable[str], resolution: float = DEFAULT_RESOLUTION) -> List[CodedLocation]:
    """
    Get the coded locations from a list of identifiers.

    Identifiers can be any combination of:
        - a location string (latitude~longitude)
        - location list (key in nzhsm_common.location.location.LOCATION_LISTS)
        - location code (e.g. "WLG")
        - grid name in nzshm_common.grids.region_grid.RegionGrid
        - csv file with at least a column headed "lat" and a column headed "lon" (any other columns will be ignored)

    Parameters:
        locations: a list of location identifiers
        resolution: the resolution used by CodedLocation

    Returns:
        coded_locations: a list of coded locations
    """
    coded_locations: List[CodedLocation] = []
    for loc_id in locations:
        location_id = str(loc_id)
        if Path(location_id).exists():
            coded_locations += _load_csv(Path(location_id), resolution)
        elif '~' in location_id:
            lat, lon = location_id.split('~')
            coded_locations.append(CodedLocation(float(lat), float(lon), resolution))
        elif location_by_id(location_id):
            coded_locations.append(CodedLocation(*_lat_lon(location_id), resolution))  # type: ignore
        elif LOCATION_LISTS.get(location_id):
            location_ids = LOCATION_LISTS[location_id]["locations"]
            coded_locations += [CodedLocation(*_lat_lon(_id), resolution) for _id in location_ids]  # type: ignore
        else:
            try:
                coded_locations += [CodedLocation(*loc, resolution) for loc in load_grid(location_id)]  # type: ignore
            except KeyError:
                msg = "location {} is not a valid location identifier".format(location_id)
                raise KeyError(msg)

    return coded_locations


def get_location_list_names() -> List[str]:
    """
    Return a list of valid location lists.

    Examples:
        >>> from nzshm_common import location
        >>> location.get_location_list_names()
        ['HB', 'NZ', 'NZ2', 'SRWG214', 'ALL']
    """
    return list(LOCATION_LISTS.keys())


def get_location_list(
    location_list_names: List[str], resolution: float = DEFAULT_RESOLUTION, sort_locations: bool = True
) -> Iterable[CodedLocation]:
    """
    Get all coded locations within one or more lists.

    The sorting method used for CodedLocation values is described in
    [`CodedLocation.__lt__`](coded_location.md#nzshm_common.location.coded_location.CodedLocation.__lt__)

    Parameters:
        location_list_names: a list of valid LOCATION_LIST keys
        resolution: the resolution used by CodedLocation
        sort_locations: (optional) whether to sort the CodedLocation values

    Returns:
        a list of coded locations

    Examples:
        >>> from nzshm_common import location
        >>> location.get_location_list(["NZ", "SRWG214"])
        [
            CodedLocation(lat=-35.109, lon=173.262, resolution=0.001),
            CodedLocation(lat=-35.22, lon=173.97, resolution=0.001),
            ...
        ]
    """
    # Merge location lists, ensuring unique keys.
    location_keyset = set([loc for name in location_list_names for loc in LOCATION_LISTS[name]["locations"]])
    coded_locations = [
        CodedLocation(
            lat=LOCATIONS_BY_ID[loc]["latitude"],
            lon=LOCATIONS_BY_ID[loc]["longitude"],
            resolution=resolution,
        )
        for loc in location_keyset
    ]
    if sort_locations:
        return sorted(coded_locations)
    else:
        return coded_locations


if __name__ == "__main__":
    """Print all locations."""
    print("custom_site_id,lon,lat")
    for loc in LOCATIONS:
        print(f"{loc['id']},{loc['longitude']},{loc['latitude']}")



"""The origional design of CodedLocation caused deserialization by pydantic BaseModel to not initialize
correctly and be missing the _code attribute.

https://github.com/GNS-Science/nzshm-common-py/issues/76

"""

from dataclasses import dataclass

import pytest
from pydantic import BaseModel

from nzshm_common import CodedLocation


def test_deserialize_pydantic_bug():
    """This test reproduces the origional fault for documentation purposes."""

    @dataclass(init=False)
    class DummyCodedLocation:
        a: int

        def __init__(self, a):
            self.a = 1
            self.b = str(self.a)

    class DummyMyModel(BaseModel):
        location: DummyCodedLocation

    loc = DummyCodedLocation(a=1)
    model = DummyMyModel(location=loc)
    data = model.model_dump()
    model_deser = DummyMyModel(**data)

    with pytest.raises(AttributeError):
        model_deser.b


def test_deserialize_pydantic_76():
    """Test that we can deserialize a BaseClass with a CodedLocation member."""

    class MyModel(BaseModel):
        location: CodedLocation

    loc = CodedLocation(-45.27, 175.2, 0.1)
    model = MyModel(location=loc)
    data = model.model_dump()

    model_deser = MyModel(**data)
    assert model.location == model_deser.location
    assert model_deser.location._code
    assert model_deser.location.grid_res
    assert model_deser.location.display_places



"""
This module contains common type definitions.

When imported from an external package, the types defined here should also be
available at the top level of the package, e.g.:

    >>> from nzshm_common import LatLon
"""

from typing import NamedTuple


class LatLon(NamedTuple):
    """
    A lightweight type for `(latitude, longitude)` float pairs.

    This is a named tuple with latitude and longitude fields.

    Examples:
        ```py
        >>> wlg = LatLon(-41.3, 174.78)
        >>> wlg
        LatLon(latitude=-41.3, longitude=174.78)
        >>> wlg.latitude
        -41.3
        >>> wlg[0]
        -41.3
        ```
    """

    latitude: float
    longitude: float


LatLon.__module__ = "nzshm_common"



"""
This package provides classes and functions to handle geographical locations. There are a number of
pre-set locations of interest to the NZ NSHM made availabe in this package. They can be accessed via
the "location lists." See examples below.

Classes:
 CodedLocation: a location defined by a latitude, longitude pair at a given resolution
 CodedLocationBin: a collection of CodedLocations bined at a given resolution (genrally lower than
 the resolution of the CodedLocations themselves)

Functions:
 get_location_list_names: get the names of the "location lists" which are lists of pre-set locations
 get_location_list: get all locations from one or more location lists as CodedLocations
 commonly used in the analysis of the NZ NSHM.
 get_locations: convert a variety of location identifiers into an iterable of CodedLocations
 location_by_id: get information about a particular location in any of the "location lists"

Example:
    ```py
    >>> all_locations = get_location_list(get_location_list_names())
    >>> nz_cities = get_location_list(["NZ"])
    >>> wellington = location_by_id("WLG")
    >>> locs = get_locations(["WLG", "CHC"])
    ```
"""

from .coded_location import CodedLocation, CodedLocationBin
from .location import get_location_list, get_location_list_names, get_locations, get_name_with_macrons, location_by_id
